<?php
// UNB Components
// Copyright since 2007 by Yves Goergen
// Website: http://unclassified.de/go/unb2
// Licence: GNU General Public License (GPL) Version 3
//
// This file contains the UnbLog class.

if (!class_exists('UnbDatabase', false)) require_once dirname(__FILE__) . '/database.class.php';
if (!class_exists('UnbSession', false)) require_once dirname(__FILE__) . '/session.class.php';
if (!class_exists('UnbEnvironment', false)) require_once dirname(__FILE__) . '/environment.class.php';

// Enumeration of event types.
//
class UnbLogType
{
	// Information log entry.
	const Information = 1;
	// Warning log entry.
	const Warning = 2;
	// Error log entry.
	const Error = 3;
	// Security-related log entry.
	const Security = 4;
	// Debug message log entry.
	const Debug = 5;
} // class UnbLogType

// Enumeration of event IDs.
//
class UnbLogEvent
{
	// A user has successfully logged in.
	const UserLogin = 1;
	// A user has logged out.
	const UserLogout = 2;
	// A user has tried to log in and failed.
	const UserLoginFailed = 3;
} // class UnbLogEvent

// The tag class provides methods for logging events and accessing the log events.
//
// It depends on the UnbDatabase class for a database connection, UnbSession for session user
// recognition and UnbEnvironment for client-specific information.
//
class UnbLog
{
	// Separate database connection
	private static $db = null;

	// Private constructor, does nothing but preventing an unwanted instantiation of this class.
	//
	private function __construct()
	{
	}

	// Gets the separate database connection used for logging
	//
	private static function GetDatabase()
	{
		if (!isset(self::$db))
		{
			self::$db = UnbDatabase::GetNewInstance();
			$connectData = array();

			// Read common database connection settings
			$entries = UnbConfiguration::Get('database.*');
			foreach ($entries as $name => $data)
			{
				$connectData[substr($name, 9)] = $data;
			}

			// Concurrent connections cannot be persistent by default, so turn it off
			$connectData['persistent'] = false;
			// Preset a different filename for SQLite
			if ($connectData['driver'] == 'sqlite')
			{
				$connectData['filename'] = $connectData['filename'] . '.log.db';
			}

			// Overwrite with log-specific settings
			$entries = UnbConfiguration::Get('database.log.*');
			foreach ($entries as $name => $data)
			{
				$connectData[substr($name, 13)] = $data;
			}

			self::$db->Open($connectData);
		}
		return self::$db;
	}

	// Adds a new log entry.
	//
	// type = (int) Event type. See the UnbLogType constants.
	// id = (int) Event ID. See the UnbLogEvent constants.
	// data = (array) Associative array of event data.
	// overrideUserId = (int) Use this user ID for the log instead of the session user.
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbDatabaseException
	//
	public static function Add($type, $id, $data = null, $overrideUserId = null)
	{
		if (!is_int($type))
			throw new InvalidArgumentException('Invalid argument type: $type');
		if ($type <= 0)
			throw new InvalidArgumentException('Argument out of range: $type');
		if (!is_int($id))
			throw new InvalidArgumentException('Invalid argument type: $id');
		if ($id <= 0)
			throw new InvalidArgumentException('Argument out of range: $id');
		if (isset($data))
		{
			if (!is_array($data))
				throw new InvalidArgumentException('Invalid argument type: $data');
			if (!count($data))
				$data = null;
		}
		if (isset($overrideUserId))
		{
			if (!is_int($overrideUserId))
				throw new InvalidArgumentException('Invalid argument type: $overrideUserId');
			if ($overrideUserId < 0)
				throw new InvalidArgumentException('Argument out of range: $overrideUserId');
		}

		// Determine the current user ID.
		// Use a user ID provided with the data in case the event is about a failed login event or such.
		if (isset($overrideUserId))
		{
			$userId = $overrideUserId;
		}
		else
		{
			$userId = UnbSession::GetSessionUserId();
		}

		$db = self::GetDatabase();
		$db->StartTransaction();
		try
		{
			$db->Execute(
				'INSERT INTO "#log" ("Num", "Time", "Type", "EventId", "UserId", "ClientAddress", "Data")
					VALUES (?, ?, ?, ?, ?, ?, ?)',
				array(new UnbDbAutoValue('log'), UnbDatabase::ConvertDateTime(time()), $type, $id, $userId, UnbEnvironment::GetClientAddress(), $data));

			$db->Commit();
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}

		// While we're here, see if we can purge old log entries
		if (mt_rand(0, 100) <= 3)
		{
			self::Purge();
		}
	}

	// Finds log entries by their event number.
	//
	// eventNumMin = (int) First event number to include in the result
	// eventNumMax = (int) Last event number to include in the result, if specified
	//
	// Throws UnbSecurityException if not logged in or no administrator.
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbDatabaseException
	//
	// Returns: (array) Array of log entries.
	//
	// Privileges: Must be logged in with administrator privilege
	//
	public static function FindByEventNum($eventNumMin, $eventNumMax = null)
	{
		if (!is_int($eventNumMin))
			throw new InvalidArgumentException('Invalid argument type: $eventNumMin');
		if ($eventNumMin < 0)
			throw new InvalidArgumentException('Argument out of range: $eventNumMin');
		if (isset($eventNumMax))
		{
			if (!is_int($eventNumMax))
				throw new InvalidArgumentException('Invalid argument type: $eventNumMax');
			if ($eventNumMax < $eventNumMin)
				throw new InvalidArgumentException('Argument out of range: $eventNumMax');
		}

		if (!UnbSession::IsAdministrator())
			throw new UnbSecurityException('You need the administrator privilege to delete log entries.');

		$sql = 'SELECT * FROM "#log" WHERE "Num" >= ?';
		$params = array($eventNumMin);
		if (isset($eventNumMax))
		{
			$sql .= ' AND "Num" <= ?';
			$params[] = $eventNumMax;
		}
		$sql .= ' ORDER BY "Num"';

		$db = self::GetDatabase();
		$rec = $db->Query($sql, $params);
		return $rec;
	}

	// Finds log entries by their properties. For internal use only.
	//
	// filter = (array) Array of filter parameters - TODO: documentation
	// countOnly = (int) true to count the events, false to return their full data
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbDatabaseException
	//
	// Returns: (array) Array of log entries.
	//
	private static function FindByFilterInternal($filter, $countOnly)
	{
		if (!is_array($filter))
			throw new InvalidArgumentException('Invalid argument type: $filter');

		$where = '';
		$params = array();

		if (isset($filter['time-notbefore']))
		{
			if (!is_int($filter['time-notbefore']))
				throw new InvalidArgumentException('Invalid argument type: $filter[\'time-notbefore\']');
			$where .= ($where ? ' AND ' : '') . '"Time" >= ?';
			$params[] = UnbDatabase::ConvertDateTime($filter['time-notbefore']);
		}
		if (isset($filter['time-notafter']))
		{
			if (!is_int($filter['time-notafter']))
				throw new InvalidArgumentException('Invalid argument type: $filter[\'time-notafter\']');
			$where .= ($where ? ' AND ' : '') . '"Time" <= ?';
			$params[] = UnbDatabase::ConvertDateTime($filter['time-notafter']);
		}
		if (isset($filter['type']))
		{
			if (is_array($filter['type']))
			{
				$where .= ($where ? ' AND ' : '') . '"Type" IN (';
				$c = 0;
				foreach ($filter['type'] as $type)
				{
					if (!is_int($type))
						throw new InvalidArgumentException('Invalid argument type: $filter[\'type\'][...]');
					if ($type <= 0)
						throw new InvalidArgumentException('Argument out of range: $filter[\'type\'][...]');
					$where .= ($c ? ', ' : '') . '?';
					$params[] = $type;
					$c++;
				}
				$where .= ')';
			}
			else
			{
				if (!is_int($filter['type']))
					throw new InvalidArgumentException('Invalid argument type: $filter[\'type\']');
				if ($filter['type'] <= 0)
					throw new InvalidArgumentException('Argument out of range: $filter[\'type\']');
				$where .= ($where ? ' AND ' : '') . '"Type" = ?';
				$params[] = $filter['type'];
			}
		}
		if (isset($filter['id']))
		{
			if (is_array($filter['id']))
			{
				$where .= ($where ? ' AND ' : '') . '"EventId" IN (';
				$c = 0;
				foreach ($filter['id'] as $id)
				{
					if (!is_int($id))
						throw new InvalidArgumentException('Invalid argument type: $filter[\'id\'][...]');
					if ($id <= 0)
						throw new InvalidArgumentException('Argument out of range: $filter[\'id\'][...]');
					$where .= ($c ? ', ' : '') . '?';
					$params[] = $id;
					$c++;
				}
				$where .= ')';
			}
			else
			{
				if (!is_int($filter['id']))
					throw new InvalidArgumentException('Invalid argument type: $filter[\'id\']');
				if ($filter['id'] <= 0)
					throw new InvalidArgumentException('Argument out of range: $filter[\'id\']');
				$where .= ($where ? ' AND ' : '') . '"EventId" = ?';
				$params[] = $filter['id'];
			}
		}
		if (isset($filter['userid']))
		{
			if (is_array($filter['userid']))
			{
				$where .= ($where ? ' AND ' : '') . '"UserId" IN (';
				$c = 0;
				foreach ($filter['userid'] as $userid)
				{
					if (!is_int($userid))
						throw new InvalidArgumentException('Invalid argument type: $filter[\'userid\'][...]');
					if ($userid <= 0)
						throw new InvalidArgumentException('Argument out of range: $filter[\'userid\'][...]');
					$where .= ($c ? ', ' : '') . '?';
					$params[] = $userid;
					$c++;
				}
				$where .= ')';
			}
			else
			{
				if (!is_int($filter['userid']))
					throw new InvalidArgumentException('Invalid argument type: $filter[\'userid\']');
				if ($filter['userid'] <= 0)
					throw new InvalidArgumentException('Argument out of range: $filter[\'userid\']');
				$where .= ($where ? ' AND ' : '') . '"UserId" = ?';
				$params[] = $filter['userid'];
			}
		}
		if (isset($filter['clientaddress']))
		{
			if (!is_string($filter['clientaddress']))
				throw new InvalidArgumentException('Invalid argument type: $filter[\'clientaddress\']');
			$where .= ($where ? ' AND ' : '') . '"ClientAddress" = ?';
			$params[] = $filter['clientaddress'];
		}

		if ($countOnly)
			$sql = 'SELECT COUNT(*)';
		else
			$sql = 'SELECT *';
		$sql .= ' FROM "#log"';
		if ($where)
			$sql .= ' WHERE ' . $where;
		if (!$countOnly)
			$sql .= ' ORDER BY "Num"';

		$db = self::GetDatabase();
		if ($countOnly)
			return $db->QueryValue($sql, $params);
		else
			return $db->Query($sql, $params);
	}

	// Finds log entries by their properties.
	//
	// filter = (array) Array of filter parameters. See FindByFilter for details.
	//
	// Throws UnbSecurityException if not logged in or no administrator.
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbDatabaseException
	//
	// Returns: (array) Array of log entries.
	//
	// Privileges: Must be logged in with administrator privilege
	//
	public static function FindByFilter($filter)
	{
		if (!UnbSession::IsAdministrator())
			throw new UnbSecurityException('You need the administrator privilege to delete log entries.');

		return self::FindByFilterInternal($filter, false);
	}

	// Counts log entries filtered by their properties.
	//
	// filter = (array) Array of filter parameters. See FindByFilter for details.
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbDatabaseException
	//
	// Returns: (int) Count of matched log entries.
	//
	public static function CountByFilter($filter)
	{
		return self::FindByFilterInternal($filter, true);
	}

	// Deletes old log entries. For internal use only.
	//
	// time = (int) Timestamp before which events will be deleted. For security reasons, the timestamp must not be newer than 7 days ago.
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbDatabaseException
	//
	private static function DeleteBeforeInternal($time)
	{
		if (!is_int($time))
			throw new InvalidArgumentException('Invalid argument type: $time');
		if ($time > time() - 7 * 86400)
			throw new InvalidArgumentException('Argument out of range: $time');

		$db = self::GetDatabase();
		$db->StartTransaction();
		try
		{
			$db->Execute('DELETE FROM "#log" WHERE "Time" < ?', UnbDatabase::ConvertDateTime($time));
			$db->Commit();
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}
	}

	// Deletes old log entries.
	//
	// time = (int) Timestamp before which events will be deleted. For security reasons, the timestamp must not be newer than 7 days ago.
	//
	// Throws UnbSecurityException if not logged in or no administrator.
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbDatabaseException
	//
	// Privileges: Must be logged in with administrator privilege
	//
	public static function DeleteBefore($time)
	{
		if (!UnbSession::IsAdministrator())
			throw new UnbSecurityException('You need the administrator privilege to delete log entries.');

		self::DeleteBeforeInternal($time);
	}

	// Deletes old log entries as specified in the configuration.
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbDatabaseException
	//
	private static function Purge()
	{
		// Make sure that no log entries younger than 7 days are deleted.
		$time = time() - max(7, intval(UnbConfiguration::Get('log.keep-days'))) * 86400;

		self::DeleteBeforeInternal($time);
	}

} // class UnbTag

?>
