<?php
// UNB Components
// Copyright since 2007 by Yves Goergen
// Website: http://unclassified.de/go/unb2
// Licence: GNU General Public License (GPL) Version 3
//
// This file contains the UnbSearch class.

if (!class_exists('UnbDatabase', false)) require_once dirname(__FILE__) . '/database.class.php';
if (!class_exists('UnbMessage', false)) require_once dirname(__FILE__) . '/message.class.php';
if (!class_exists('UnbMessageRevision', false)) require_once dirname(__FILE__) . '/messagerevision.class.php';
if (!class_exists('UnbTag', false)) require_once dirname(__FILE__) . '/tag.class.php';
if (!class_exists('UnbUser', false)) require_once dirname(__FILE__) . '/user.class.php';
if (!class_exists('UnbSession', false)) require_once dirname(__FILE__) . '/session.class.php';

// The search class provides search methods and handles search requests and their results.
//
// It depends on the UnbDatabase class for a database connection. It also uses the classes
// UnbMessage, UnbMessageRevision, UnbTag and UnbUser to handle search queries and results related
// to those item types.
//
class UnbSearch
{
	// Cached value from GetInaccessibleTagIds
	private static $inaccessibleTagIds = null;

	// Private constructor, does nothing but preventing an unwanted instantiation of this class.
	//
	private function __construct()
	{
	}

	// Gets all tag IDs that the session user has no read access to.
	//
	// Returns: (array(int)) Tag IDs
	//
	private static function GetInaccessibleTagIds()
	{
		if (!isset(self::$inaccessibleTagIds))
		{
			$sessionUserKeys = UnbSession::GetSessionUserKeys();
			if (count($sessionUserKeys))
				$userIdPredicate = 'IN (' . join(', ', $sessionUserKeys) . ')';
			else
				$userIdPredicate = '= 0';

			$sql = 'SELECT t."TagId"
				FROM "#tag" t
					LEFT JOIN "#keylist_keys" tkk ON
						(tkk."KeylistId" = t."ReadAccessKeylistId" AND
						tkk."UserId" ' . $userIdPredicate . ')
				WHERE tkk."UserId" IS NULL AND
					t."ReadAccessKeylistId" IS NOT NULL';

			self::$inaccessibleTagIds = UnbDatabase::GetInstance()->QueryArray($sql);
		}
		return self::$inaccessibleTagIds;
	}

	// Constructs the database query to perform the search.
	//
	// searchParameters = (array) Array of search parameters
	//
	// Returns: (array) SQL string at first index, parameters array at second index.
	//
	private static function ConstructSqlQuery($searchParameters)
	{
		$sessionUserId = UnbSession::GetSessionUserId();
		$isModerator = UnbSession::IsModerator();
		$isAdministrator = UnbSession::IsAdministrator();

		if (is_int($searchParameters['owner-ids']))
			$searchParameters['owner-ids'] = array($searchParameters['owner-ids']);

		$getTagData = intval($searchParameters['get-tag-data']);

		$parameters = array();
		$sql = '';
		if ($getTagData > 0)
			$sql .= 'SELECT mrtGTD."TagData" AS "TagData", COUNT(*) AS "Count"';
		else
			$sql .= 'SELECT m."MessageId"';

		$sql .= ' FROM "#message" m';

		$sortTagId = intval($searchParameters['sort-tagid']);
		if ($sortTagId > 0)
		{
			$sql .= ' JOIN "#message_revision_tag" mrt
				ON mrt."MessageRevisionId" = m."SearchRevisionId" AND mrt."TagId" = ?';
			$parameters[] = $sortTagId;
		}
		if (is_array($searchParameters['tag-value']))
		{
			foreach ($searchParameters['tag-value'] as $tagId => $value)
			{
				if (!is_numeric($tagId)) continue;

				$sql .= ' JOIN "#message_revision_tag" mrTV' . $tagId . '
					ON mrTV' . $tagId . '."MessageRevisionId" = m."SearchRevisionId" AND
						mrTV' . $tagId . '."TagId" = ?';
				$parameters[] = $tagId;
			}
		}

		if ($getTagData > 0)
		{
			$sql .= ' JOIN "#message_revision_tag" mrtGTD ON mrtGTD."MessageRevisionId" = m."SearchRevisionId" AND mrtGTD."TagId" = ?';
			$parameters[] = $getTagData;
		}

		$sql .= '
			WHERE
				-- Don\'t find drafts
				NOT m."Draft" AND
				-- Grant all access to administrators (so that they can alter the access key lists)
				(? OR
					-- Message has no read access keylist
					(m."ReadAccessKeylistId" IS NULL
					OR EXISTS
						-- Message\'s read access keylist contains personal or additional user key
						(SELECT 1
						FROM "#keylist_keys" mkk
						WHERE mkk."KeylistId" = m."ReadAccessKeylistId" AND
							(mkk."UserId" = ? OR
								mkk."UserId" IN
									(SELECT ukk."UserId"
									FROM "#user" u, "#keylist_keys" ukk
									WHERE ukk."KeylistId" = u."AdditionalKeylistId" AND
										u."UserId" = ?))))';
		$parameters[] = $isAdministrator;
		$parameters[] = $sessionUserId;
		$parameters[] = $sessionUserId;

		$inaccessibleTagIds = self::GetInaccessibleTagIds();
		if (count($inaccessibleTagIds))
		{
			$dbDriver = UnbDatabase::GetInstance()->GetDriverName();
			if ($dbDriver != 'pgsql')
			{
				$sql .= '
						AND NOT EXISTS
							-- There must not be any tag that denies access
							(SELECT "TagId"
							FROM "#message_revision_tag" mrt
								JOIN "#tag" t USING ("TagId")
							WHERE mrt."MessageRevisionId" = m."SearchRevisionId" AND
								t."TagId" IN (' . join(', ', $inaccessibleTagIds) . '))';
			}
			else
			{
				// TODO: Re-test the above code with PostgreSQL! It's likely to work now...

				// This is for PostgreSQL:
				$sql .= '
						AND NOT EXISTS
							-- There must not be any tag that denies access
							(SELECT "Allowed"
							FROM
								(SELECT
									EXISTS
										-- Tag\'s read access keylist contains personal or additional user key
										(SELECT 1
										FROM "#keylist_keys" tkk
										WHERE tkk."KeylistId" = t."ReadAccessKeylistId" AND
											(tkk."UserId" = ? OR
												tkk."UserId" IN
													(SELECT ukk."UserId"
													FROM "#user" u, "#keylist_keys" ukk
													WHERE ukk."KeylistId" = u."AdditionalKeylistId" AND
														u."UserId" = ?)))
									AS "Allowed"
								FROM "#message" m2   -- TODO: remove additional message table reference (m2 = m), see above
									JOIN "#message_revision_tag" mrt
										ON mrt."MessageRevisionId" = m2."SearchRevisionId"
									JOIN "#tag" t USING ("TagId")
								WHERE m2."MessageId" = m."MessageId" AND
									-- Only regard tags with read access keylist (only they can fail the access test)
									t."ReadAccessKeylistId" IS NOT NULL) AS foo
							WHERE NOT "Allowed")';
			}
		}
		$sql .= ')';

		if (!$isModerator && !$isAdministrator)
		{
			// Message must not be hidden - message owner is exempted
			$sql .= '
				AND (NOT (m."ModerationFlags" & ' . UnbMessageModerationFlag::Hidden . ') OR
					m."Owner" = ?)';
			$parameters[] = $sessionUserId;
		}

		// The following conditions are not access-related

		if (is_array($searchParameters['message-ids']) && count($searchParameters['message-ids']))
		{
			$messageIds = array_map('intval', $searchParameters['message-ids']);
			$sql .= '
				AND m."MessageId" IN (' . join(', ', $messageIds) . ')';
		}

		if (is_array($searchParameters['owner-ids']) && count($searchParameters['owner-ids']))
		{
			$ownerIds = array_map('intval', $searchParameters['owner-ids']);
			$sql .= '
				AND m."Owner" ' .
				($searchParameters['owner-ids-inverse'] ? 'NOT ' : '') .
				'IN (' . join(', ', $ownerIds) . ')';
		}

		if (is_bool($searchParameters['entrypoint']))
		{
			if ($searchParameters['entrypoint'] === true)
				$sql .= '
					AND m."EntryPoint"';
			else if ($searchParameters['entrypoint'] === false)
				$sql .= '
					AND NOT m."EntryPoint"';
		}

		if (is_int($searchParameters['primary-reference']))
		{
			if ($searchParameters['primary-reference'] > 0)
			{
				$sql .= '
					AND m."PrimaryReference" = ?';
				$parameters[] = $searchParameters['primary-reference'];
			}
			else
			{
				$sql .= '
					AND m."PrimaryReference" IS NULL';
			}
		}

		if (is_array($searchParameters['tags-set']) && count($searchParameters['tags-set']))
		{
			foreach ($searchParameters['tags-set'] as $tagId)
			{
				$sql .= '
					AND ' . ($tagId < 0 ? 'NOT ' : '') . 'EXISTS
						(SELECT 1
						FROM "#message_revision_tag" mrt
						WHERE mrt."MessageRevisionId" = m."SearchRevisionId" AND
							mrt."TagId" = ?)';
				$parameters[] = abs($tagId);
			}
		}

		if (is_array($searchParameters['tag-value']))
		{
			foreach ($searchParameters['tag-value'] as $tagId => $value)
			{
				if (!is_numeric($tagId)) continue;

				$sql .= ' AND mrTV' . $tagId . '."TagData" ';
				if (is_array($value))
				{
					$sql .= 'BETWEEN ? AND ?';
					$parameters[] = $value[0];
					$parameters[] = $value[1];
				}
				else if (strpos($value, '*') !== false || strpos($value, '?') !== false)
				{
					$sql .= 'LIKE ? ESCAPE ?';
					$parameters[] = UnbDatabase::EscapeLikeWildcards($value);
					$parameters[] = '\\';
				}
				else
				{
					$sql .= '= ?';
					$parameters[] = $value;
				}
			}
		}

		if ($searchParameters['createdtime-notbefore'])
		{
			$sql .= '
				AND "CreatedTime" >= ?';
			$time = $searchParameters['createdtime-notbefore'];
			if (is_int($time))
				$time = UnbDatabase::ConvertDateTime($time);
			$parameters[] = $time;
		}
		if ($searchParameters['createdtime-notafter'])
		{
			$sql .= '
				AND "CreatedTime" <= ?';
			$time = $searchParameters['createdtime-notafter'];
			if (is_int($time))
				$time = UnbDatabase::ConvertDateTime($time);
			$parameters[] = $time;
		}

		if ($getTagData > 0)
		{
			$sql .= ' GROUP BY mrtGTD."TagData"';
		}

		if ($sortTagId > 0)
		{
			$sql .= ' ORDER BY mrt."TagData"';
			if ($searchParameters['sort-reverse'])
				$sql .= ' DESC';
		}
		else if ($searchParameters['sort-createdtime'])
		{
			$sql .= ' ORDER BY "CreatedTime"';
			if ($searchParameters['sort-reverse'])
				$sql .= ' DESC';
		}

		if (preg_match('_^\d+(\s*,\s*\d+)?$_', $searchParameters['limit']))
		{
			$sql .= ' LIMIT ' . $searchParameters['limit'];
		}

		return array($sql, $parameters);
	}

	// Processes a search query.
	//
	// query = (array) Array of search parameters
	//
	// Throws UnbDatabaseException
	//
	// Returns (array(int)) Array of message IDs found by the search query
	//
	public static function Query($query = null)
	{
		$searchParameters = array();
		if (is_array($query))
			$searchParameters = $query;
		#$searchParameters['message-ids'] = array(1, 2, 3, 4, 5);
		#$searchParameters['owner-ids'] = array(1, 2);
		#$searchParameters['owner-ids-inverse'] = false;
		#$searchParameters['entrypoint'] = true;
		#$searchParameters['tags-set'] = array(21);
		#$searchParameters['tag-value'][1] = array('2005-01-01', '2009-12-31');
		#$searchParameters['primary-reference'] = 1;
		#$searchParameters['sort-tagid'] = 13;
		#$searchParameters['sort-createdtime'] = false;
		#$searchParameters['sort-reverse'] = false;
		#$searchParameters['limit'] = '';
		#$searchParameters['createdtime-notbefore'] = '2000-01-01';   // or unix timestamp
		#$searchParameters['createdtime-notafter'] = '2020-12-31';   // or unix timestamp

		list($sql, $parameters) = self::ConstructSqlQuery($searchParameters);
		$db = UnbDatabase::GetInstance();
		#$t0 = microtime(true);
		$messageIds = $db->QueryArray($sql, $parameters);
		#echo round((microtime(true) - $t0) * 1000) . 'ms ';
		if (!isset($messageIds)) return null;
		$messageIds = array_map('intval', $messageIds);

		// All messages returned from this query are accessible by the current session user.
		// To avoid any further checks, cache these message IDs.
		UnbMessage::SetReadAllowedIds($messageIds);

		return $messageIds;
	}

	// Finds all messages that directly or indirectly reference to a given message through their
	// primary references.
	//
	// messageId = (int) Base message ID, the search starts here
	//
	// Returns (array(int => int)) Array of message IDs as keys, with their direct primary reference as value
	//
	public static function GetAllReplies($messageId)
	{
		if (!is_int($messageId))
			throw new InvalidArgumentException('Invalid argument type: $messageId');
		if ($messageId <= 0)
			throw new InvalidArgumentException('Argument out of range: $messageId');

		$out = array();
		self::GetAllReplies_rec(0, $messageId, array(), $out);
		return $out;
	}

	// Recursion helper function for GetAllReplies.
	//
	private static function GetAllReplies_rec($primaryRef, $messageId, $seenMessageIds, &$out)
	{
		if ($primaryRef)
			$out[$messageId] = $primaryRef;

		if (in_array($messageId, $seenMessageIds))
		{
			// Reference loop!
			return;
		}

		$seenMessageIds[] = $messageId;

		// Find all messages referencing on this one:
		$searchParameters = array();
		$searchParameters['primary-reference'] = $messageId;
		$messageIds = UnbSearch::Query($searchParameters);

		if ($messageIds)
		{
			foreach ($messageIds as $replyMessageId)
			{
				self::GetAllReplies_rec($messageId, $replyMessageId, $seenMessageIds, $out);
			}
		}
	}

	public static function GetMessageTree($messageId, $rootMessageId = null, $autoSelect = true, $maxLevel = null)
	{
		if (!is_int($messageId))
			throw new InvalidArgumentException('Invalid argument type: $messageId');
		if ($messageId <= 0)
			throw new InvalidArgumentException('Argument out of range: $messageId');
		if (isset($rootMessageId))
		{
			if (!is_int($rootMessageId))
				throw new InvalidArgumentException('Invalid argument type: $rootMessageId');
		}
		if (!is_bool($autoSelect))
			throw new InvalidArgumentException('Invalid argument type: $autoSelect');
		if (isset($maxLevel))
		{
			if (!is_int($maxLevel))
				throw new InvalidArgumentException('Invalid argument type: $maxLevel');
		}

		// Follow all messages' primary reference up to the beginning of the discussion
		$msgCache = array();
		$revCache = array();

		$msg = UnbMessage::FindByMessageId($messageId);
		if (!isset($msg))
			throw new Exception('The specified message ID does not exist.');
		$rev = $msg->GetSearchRevision(UnbMessageRevisionContentLevel::HtmlContent, true);

		$msgCache[$msg->GetMessageId()] = $msg;
		$revCache[$rev->GetMessageRevisionId()] = $rev;

		while (!$msg->IsEntryPoint() && $msg->GetPrimaryReference())
		{
			if ($msg->GetMessageId() == $rootMessageId) break;   // Don't read over the root message, even if it's not the DEP yet

			$msg = UnbMessage::FindByMessageId($msg->GetPrimaryReference());
			if (!isset($msg))
				throw new Exception('The message ID ' . $msg->GetPrimaryReference() . ' does not exist (database integrity error).');
			$rev = $msg->GetSearchRevision(
				$autoSelect ? UnbMessageRevisionContentLevel::HtmlContent : UnbMessageRevisionContentLevel::Subject,
				true);

			$msgCache[$msg->GetMessageId()] = $msg;
			$revCache[$rev->GetMessageRevisionId()] = $rev;
		}

		if (!isset($rootMessageId))
		{
			$rootMessageId = $msg->GetMessageId();
		}

		$seenMessageIds = array();
		$out = array();
		self::GetMessageTree_rec($rootMessageId, $messageId, $autoSelect, 0, $maxLevel, 0, $seenMessageIds, $msgCache, $revCache, $out);
		return $out;
	}

	private static function GetMessageTree_rec($messageId, $selectedMessageId, $autoSelect, $level, $maxLevel, $index, &$seenMessageIds, &$msgCache, &$revCache, &$out)
	{
		if (in_array($messageId, $seenMessageIds)) return;
		$seenMessageIds[] = $messageId;

		if ($autoSelect)
		{
			// Automatically select all messages up-path the originally specified message and all
			// first replies in each level below.
			$selected = false;
			if (array_key_exists($messageId, $msgCache)) $selected = true;   // Message was seen up-path
			if ($level >= count($msgCache))
			{
				// We're now below the specified level, now select each first reply
				if ($index == 0) $selected = true;
			}
		}
		else
		{
			// Only select the originally specified message.
			$selected = $messageId == $selectedMessageId;
		}

		$msg = $msgCache[$messageId];
		if (!isset($msg))
			$msg = UnbMessage::FindByMessageId($messageId);
		if (!isset($msg))
			throw new Exception('The message ID ' . $messageId . ' does not exist (database integrity error).');
		$rev = $revCache[$msg->GetSearchRevisionId()];
		if (!isset($rev))
			$rev = $msg->GetSearchRevision(
				$selected ? UnbMessageRevisionContentLevel::HtmlContent : UnbMessageRevisionContentLevel::Subject,
				true);

		if ($level <= $maxLevel)
		{
			$item = array();
			$item['level'] = $level;
			$item['msg'] =& $msg;
			$item['rev'] =& $rev;
			$item['selected'] = $selected;
			#if ($level >= $maxLevel)
			#{
			#	$item['morelevels'] = true;
			#}
			$out[] = $item;
		}
		else
		{
			$out[count($out) - 1]['morecount']++;
		}

		#if ($level < $maxLevel)
		{
			// Find all messages referencing on this one:
			$searchParameters = array();
			$searchParameters['sort-createdtime'] = true;
			$searchParameters['primary-reference'] = $messageId;
			$messageIds = UnbSearch::Query($searchParameters);

			$subIndex = 0;
			foreach ($messageIds as $messageId)
			{
				self::GetMessageTree_rec(
					$messageId,
					$selectedMessageId,
					$autoSelect && $selected,   // Only keep autoSelect if we're still in the "selected path"
					$level + 1,
					$maxLevel,
					$subIndex++,
					$seenMessageIds,
					$msgCache,
					$revCache,
					$out);
			}
		}
	}

	// Finds all values of a tag with user data.
	//
	// Only messages are considered with a revision accessible to the current session user.
	//
	// query = (array) Array of search parameters
	//
	// Throws UnbDatabaseException
	//
	// Returns (array(string => int)) Array of values (as key) with their ocurrence count
	//
	// TODO: What does this function and what is it good for?
	//
	public static function GetAllTagValues($tagId, $query = null)
	{
		$searchParameters = array();
		if (is_array($query))
			$searchParameters = $query;
		$searchParameters['get-tag-data'] = $tagId;
		$searchParameters['sort-tagid'] = $tagId;

		list($sql, $parameters) = self::ConstructSqlQuery($searchParameters);
		$db = UnbDatabase::GetInstance();
		$records = $db->Query($sql, $parameters);
		if (!isset($records)) return null;
		$a = array();
		foreach ($records as $rec)
			$a[$rec['TagData']] = intval($rec['Count']);
		return $a;
	}

} // class UnbSearch

?>
