<?php
// UNB Components
// Copyright since 2007 by Yves Goergen
// Website: http://unclassified.de/go/unb2
// Licence: GNU General Public License (GPL) Version 3
//
// This file contains the UnbSession class.

if (!class_exists('UnbUser', false)) require_once dirname(__FILE__) . '/user.class.php';
if (!class_exists('UnbConfiguration', false)) require_once dirname(__FILE__) . '/configuration.class.php';
if (!class_exists('UnbSecurityException', false)) require_once dirname(__FILE__) . '/unb.class.php';
if (!class_exists('UnbLog', false)) require_once dirname(__FILE__) . '/log.class.php';

// The session class manages a user session and thus logically connects multiple web page requests.
// It can establish, configure and terminate sessions as well as generate and validate login
// information.
//
// UnbSession is a static class, there is no instance of it. All methods are static, so that
// session information can be used from multiple classes without passing a reference everywhere.
//
// It depends on the UnbUser class for user account related tasks.
//
class UnbSession
{
	// Session name. Set to '' to use an automatic name. See PHP's session_name() for details.
	const SessionName = '';
	// Session IP address netmask. Set to 0 to disable this security check.
	// TODO: Support IPv6
	const SessionIpNetMask = 0xFFFFFF00;
	// Session expiration timeout in minutes. Set to 0 to disable this security check.
	const SessionExpireTimeout = 60;
	// Session cookie domain. Set to '' to use the default value.
	const CookieDomain = '';
	// Session cookie path. Set to '' to use the URL path of the calling script.
	const CookiePath = '';
	// Name of the array in $_SESSION where all UNB session data is stored into
	const SessionArrayVar = 'UNB';
	// Default system encryption key. Set a different key via the configuration setting "encryption.key".
	// At maximum 32 bytes are used as key. Any additional length is ignored and does not enhance security.
	// TODO: Make this a variable and use some constant system information for the default key.
	const DefaultEncryptionKey = 'UNB2_Key_1234567890';

	// Currently active session user (UnbUser object)
	private static $sessionUser = null;

	// Private constructor, does nothing but preventing an instantiation of this class.
	//
	private function __construct()
	{
	}

	// Initialises required internal session variables.
	//
	private static function InitialiseSession()
	{
		$_SESSION[self::SessionArrayVar] = array();
		$_SESSION[self::SessionArrayVar]['IpAddress'] = ip2long($_SERVER['REMOTE_ADDR']);   // TODO: Support IPv6
		$_SESSION[self::SessionArrayVar]['HeartBeat'] = time();
	}

	// Gets the cookie domain to be used.
	//
	// Returns (string) Cookie domain
	//
	private static function GetCookieDomain()
	{
		$cookieDomain = trim(self::CookieDomain);
		// Ensure a leading point (PHP manual says it's more compatible with browsers)
		if (strlen($cookieDomain) && $cookieDomain{0} != '.')
			$cookieDomain = '.' . $cookieDomain;
		return $cookieDomain;
	}

	// Gets the cookie path to be used.
	//
	// Returns (string) Cookie path
	//
	private static function GetCookiePath()
	{
		$cookiePath = trim(self::CookiePath);
		if (!strlen($cookiePath))
		{
			// Use the URL path of the calling script as cookie path.
			$cookiePath = dirname($_SERVER['PHP_SELF']);
		}
		// Ensure a trailing slash (required for HTTP)
		if (substr($cookiePath, -1) != '/')
			$cookiePath .= '/';
		return $cookiePath;
	}

	// Does everything required to use PHP's session support.
	//
	// sessionId = (string) PHP session ID, if not found in the request variables. Useful for non-browser environments.
	//
	// Throws InvalidArgumentException on an invalid session ID value.
	//
	private static function StartPhpSession($sessionId = null)
	{
		// Start the PHP session if none is started yet
		if (!session_id() || isset($sessionId))
		{
			$sessionname = self::SessionName;
			if (!is_string($sessionname) || trim($sessionname) === '')
			{
				// Generate a hash value of the absolute path to this source file. This ensures a
				// unique session name also for multiple application instances within the same
				// session cookie path. Only use half of the MD5 hash to save cookie space.
				$sessionname = 'unb_session_' . substr(md5(dirname(__FILE__)), 0, 16);
			}

			// Check session ID for invalid characters
			$sid = $_REQUEST[$sessionname];
			if (!preg_match('/[a-f0-9-,]{0,42}/i', $sid))
				throw new InvalidArgumentException('Invalid session ID value');

			ini_set('url_rewriter.tags', '');
			ini_set('session.use_trans_sid', '0');
			#ini_set('session.cookie_domain', self::GetCookieDomain());
			#ini_set('session.cookie_path', self::GetCookiePath());
			session_set_cookie_params(0, self::GetCookiePath(), self::GetCookieDomain());
			session_name($sessionname);
			if (isset($sessionId))
				session_id($sessionId);
			session_start();

			// Internet Explorer SSL fix
			@header('Pragma: ');
			@header('Cache-Control: cache');
		}
	}

	// Enables using a session.
	//
	// This method is called at the beginning of an application to start using a session. This
	// enables access to session variables and allows storing temporary data over multiple web page
	// requests.
	//
	// sessionId = (string) PHP session ID, if not found in the request variables. Useful for non-browser environments.
	//
	// Throws InvalidArgumentException on an invalid SessionArrayVar value.
	//
	public static function EnableSession($sessionId = null)
	{
		if (!is_string(self::SessionArrayVar) || trim(self::SessionArrayVar) === '')
			throw new InvalidArgumentException('Invalid argument type: SessionArrayVar');

		self::StartPhpSession($sessionId);

		// Initialise session data store.
		// For compatibility with other applications, all session data is stored in this single array.
		if (!isset($_SESSION[self::SessionArrayVar]))
		{
			self::InitialiseSession();
		}

		// Compare IP address
		// TODO: Support IPv6
		$ip = ip2long($_SERVER['REMOTE_ADDR']);
		if ($ip & self::SessionIpNetMask != $_SESSION[self::SessionArrayVar]['IpAddress'] & self::SessionIpNetMask)
		{
			// The IP address has changed outside of the netmask.
			// This means that the session was probably hijacked by a third party by stealing the
			// session ID value. For security, we'll now erase the session data and set the new
			// IP address. This also logs out a currently logged in user and thus drops all
			// privileges.
			self::InitialiseSession();
		}

		// Expire session on inactivity
		if (self::SessionExpireTimeout > 0 &&
			$_SESSION[self::SessionArrayVar]['HeartBeat'] < time() - self::SessionExpireTimeout * 60)
		{
			// The session was not used within the expiration timeout.
			// For security, we'll now erase the session data. This also logs out a currently
			// logged in user and thus drops all privileges.
			self::InitialiseSession();
		}
		$_SESSION[self::SessionArrayVar]['HeartBeat'] = time();

		// Load session user object
		if (isset($_SESSION[self::SessionArrayVar]['SessionUserId']))
		{
			// Check data
			if (!is_int($_SESSION[self::SessionArrayVar]['SessionUserId']))
			{
				if (is_numeric($_SESSION[self::SessionArrayVar]['SessionUserId']))
				{
					$_SESSION[self::SessionArrayVar]['SessionUserId'] = intval($_SESSION[self::SessionArrayVar]['SessionUserId']);
				}
				else
				{
					$_SESSION[self::SessionArrayVar]['SessionUserId'] = null;
				}
			}
			// Load user object from the database
			if ($_SESSION[self::SessionArrayVar]['SessionUserId'] > 0)
			{
				try
				{
					self::$sessionUser = UnbUser::FindByUserId($_SESSION[self::SessionArrayVar]['SessionUserId']);
				}
				catch (Exception $ex)
				{
					self::$sessionUser = null;
				}
			}
			if (!isset(self::$sessionUser))
			{
				$_SESSION[self::SessionArrayVar]['SessionUserId'] = null;
			}
		}

		// Auto-login from cookie
		if (!isset(self::$sessionUser))
		{
			list($userId, $pCookie) = explode(' ', $_COOKIE['unb_autologin']);
			if (isset($userId) && strlen(trim($pCookie)) > 0)
			{
				$userId = intval($userId);
				try
				{
					$user = UnbUser::FindByUserId($userId);
					if (isset($user))
					{
						$pStored = self::Decrypt($user->GetSingleSetting('autologin.password'));
						if ($pCookie !== $pStored)
							throw new Exception();
						self::SetSessionUser($user, null, true);
					}
				}
				catch (Exception $ex)
				{
				}
			}
		}
	}

	// Deletes an active session.
	//
	// Stops using an active session and deletes any data from it so that it cannot be accidently
	// reused later.
	//
	public static function DeleteSession()
	{
		StartPhpSession();
		$_SESSION[self::SessionArrayVar] = null;
		$_SESSION = array();
		session_destroy();
	}

	// Sets a variable in the session variable space.
	//
	// name = (string) Variable name
	// value = (any) Variable value, {{null}} to unset the variable
	//
	public static function SetVar($name, $value = null)
	{
		$_SESSION[self::SessionArrayVar]['Var'][$name] = $value;
	}

	// Gets a variable from the session variable space.
	//
	// name = (string) Variable name
	// unset = (bool) Unset the variable when fetched
	//
	public static function GetVar($name, $unset = false)
	{
		$value = $_SESSION[self::SessionArrayVar]['Var'][$name];
		if ($unset)
			$_SESSION[self::SessionArrayVar]['Var'][$name] = null;
		return $value;
	}

	// Sends a cookie to the browser.
	//
	// name = (string) Cookie name
	// value = (string or int) Cookie data, or {{null}} or empty to delete the cookie
	// expire = (int) UNIX timestamp of the time to expire the cookie, or 0 for a session cookie, or -1 for infinite lifetime
	//
	public static function SetCookie($name, $value = null, $expire = -1)
	{
		if (!is_string($name))
			throw new InvalidArgumentException('Invalid argument type: $name');
		$name = trim($name);
		if (!strlen($name))
			throw new InvalidArgumentException('Name is empty');
		if (!isset($value))
			$value = '';
		if (is_int($value))
			$value .= '';
		if (!is_string($value))
			throw new InvalidArgumentException('Invalid argument type: $value');
		$value = trim($value);
		if (!is_int($expire))
			throw new InvalidArgumentException('Invalid argument type: $expire');

		if (!strlen($value))
			$expire = 1;   // 1970-01-01T00:00:01+0000, definitely in the past...
		else if ($expire == -1)
			$expire = intval(pow(2, 31) - 1);   // Int32.MaxValue, somewhere around the year 2038...
		setcookie($name, $value, $expire, self::GetCookiePath(), self::GetCookieDomain());
	}

	// Gets the currently active session user.
	//
	// Throws BadMethodCallException if session not enabled.
	//
	// Returns a UnbUser object of the session user, or {{null}} if no registered session user is active.
	//
	public static function GetSessionUser()
	{
		if (!session_id())
			throw new BadMethodCallException('Session not enabled');

		return self::$sessionUser;
	}

	// Gets the ID of the currently active session user.
	//
	// Throws BadMethodCallException if session not enabled.
	//
	// Returns a positive integer of the user ID, or 0 if no registered session user is active.
	//
	public static function GetSessionUserId()
	{
		if (!session_id())
			throw new BadMethodCallException('Session not enabled');

		if (isset(self::$sessionUser)) return self::$sessionUser->GetUserId();
		return 0;
	}

	// Gets all the session user's keys, that is the user ID and all additional keys.
	//
	// Throws BadMethodCallException if session not enabled.
	//
	// Returns an array(int) of key IDs or an empty array if no registered session user is active.
	//
	public static function GetSessionUserKeys()
	{
		if (!session_id())
			throw new BadMethodCallException('Session not enabled');

		if (isset(self::$sessionUser))
		{
			return array_unique(array_merge(
				array(self::$sessionUser->GetUserId()),
				self::$sessionUser->GetAdditionalKeys()));
		}
		return array();
	}

	// Gets a value indicating whether the active session user is an administrator.
	//
	// Throws BadMethodCallException if session not enabled.
	//
	// Returns (bool)
	//
	public static function IsAdministrator()
	{
		if (!session_id())
			throw new BadMethodCallException('Session not enabled');

		if (!isset(self::$sessionUser)) return false;
		return self::$sessionUser->IsAdministrator();
	}

	// Gets a value indicating whether the active session user is a moderator.
	//
	// Throws BadMethodCallException if session not enabled.
	//
	// Returns (bool)
	//
	public static function IsModerator()
	{
		if (!session_id())
			throw new BadMethodCallException('Session not enabled');

		if (!isset(self::$sessionUser)) return false;
		return self::$sessionUser->IsModerator();
	}

	// Gets a value indicating whether the active session user is called "trusted".
	//
	// Throws BadMethodCallException if session not enabled.
	//
	// Returns (bool)
	//
	public static function IsTrusted()
	{
		if (!session_id())
			throw new BadMethodCallException('Session not enabled');

		if (!isset(self::$sessionUser)) return false;
		return self::$sessionUser->IsTrusted();
	}

	// Gets a value indicating whether the active session user is logged in.
	//
	// Throws BadMethodCallException if session not enabled.
	//
	// Returns (bool)
	//
	public static function IsLoggedIn()
	{
		if (!session_id())
			throw new BadMethodCallException('Session not enabled');

		return isset(self::$sessionUser);
	}

	// Sets a new active session user.
	//
	// sessionUser = (UnbUser) New session user. Set to {{null}} to "log out" an active user.
	// password = (string) Plain login password to validate.
	// overridePassword = (bool) Do not validate the provided password.
	//                           This is meant to be used for setup and maintenance only!
	//                           It may also be used carefully when logging in a newly created user.
	//                           When currently logged in as administrator and providing no or an
	//                           empty password, the new user's password is not validated either.
	//
	// Throws BadMethodCallException if session not enabled.
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbSecurityException on missing or rejected login password.
	//
	public static function SetSessionUser($sessionUser, $password = null, $overridePassword = false)
	{
		if (!session_id())
			throw new BadMethodCallException('Session not enabled');

		if (isset($sessionUser))
		{
			if (!($sessionUser instanceof UnbUser))
				throw new InvalidArgumentException('Invalid argument type: $sessionUser');

			if ((!isset($password) || $password === '') && self::IsAdministrator() ||
				$overridePassword)
			{
				// We're logged in as administrator and haven't provided a login password,
				// or the overridePassword argument is set true,
				// no need to validate the password.
			}
			else
			{
				if (!is_string($password))
					throw new InvalidArgumentException('Invalid argument type: $password');
				$password = trim($password);

				$sessionUserPassword = $sessionUser->GetLoginPassword();
				if (!isset($sessionUserPassword))
					throw new UnbSecurityException('[UNB#12] You cannot log in without a password.');

				// Check the log to find previous failed login events and delay this login request.
				$delayTrigger =
					// Failed login events on the entire system - 20 events in 1 minute
					UnbLog::CountByFilter(array(
						'id' => UnbLogEvent::UserLoginFailed,
						'time-notbefore' => time() - 1 * 3600,
						)) > 20 ||
					// Failed login events for the same user - 5 events in 5 minutes
					UnbLog::CountByFilter(array(
						'id' => UnbLogEvent::UserLoginFailed,
						'time-notbefore' => time() - 5 * 3600,
						'userid' => $sessionUser->GetUserId(),
						)) > 5 ||
					// Failed login events from the same client - 5 events in 5 minutes
					UnbLog::CountByFilter(array(
						'id' => UnbLogEvent::UserLoginFailed,
						'time-notbefore' => time() - 5 * 3600,
						'clientaddress' => UnbEnvironment::GetClientAddress(),
						)) > 5 ||
					// Failed login events from the same client for the same user - 3 events in 5 minutes
					UnbLog::CountByFilter(array(
						'id' => UnbLogEvent::UserLoginFailed,
						'time-notbefore' => time() - 5 * 3600,
						'clientaddress' => UnbEnvironment::GetClientAddress(),
						'userid' => $sessionUser->GetUserId(),
						)) > 3;
				if ($delayTrigger)
				{
					// Wait some time between 3 and 6 seconds to delay the login process and
					// decrease the login retry rate.
					// Wait always, no matter what the login result will be.
					// Wait a random amount of time to obscure any code runtime differences for
					// either login result.
					usleep(mt_rand(3000000, 6000000));
				}

				// Check the log to find previous failed login events and fail this login request.
				//
				// If a certain rate of failed logins is exceeded, an attack may be assumed. To
				// protect the system accounts, no brute-force attack is accepted and all login
				// requests - legitimate or not - will be denied in that situation. The ban is
				// automatically removed by waiting a while and retrying then with correct login
				// credentials.
				$failTrigger =
					// Failed login events for the same user - 50 events in 10 minutes
					UnbLog::CountByFilter(array(
						'id' => UnbLogEvent::UserLoginFailed,
						'time-notbefore' => time() - 10 * 3600,
						'userid' => $sessionUser->GetUserId(),
						)) > 50 ||
					// Failed login events from the same client - 30 events in 10 minutes
					UnbLog::CountByFilter(array(
						'id' => UnbLogEvent::UserLoginFailed,
						'time-notbefore' => time() - 10 * 3600,
						'clientaddress' => UnbEnvironment::GetClientAddress(),
						)) > 30 ||
					// Failed login events from the same client for the same user - 25 events in 10 minutes
					UnbLog::CountByFilter(array(
						'id' => UnbLogEvent::UserLoginFailed,
						'time-notbefore' => time() - 10 * 3600,
						'clientaddress' => UnbEnvironment::GetClientAddress(),
						'userid' => $sessionUser->GetUserId(),
						)) > 25;

				if ($failTrigger || !self::ValidatePassword($password, $sessionUserPassword))
				{
					UnbLog::Add(UnbLogType::Information, UnbLogEvent::UserLoginFailed, null, $sessionUser->GetUserId());
					throw new UnbSecurityException('[UNB#13] The login password you provided was rejected.');
				}
			}
		}
		else
		{
			UnbLog::Add(UnbLogType::Information, UnbLogEvent::UserLogout);
		}

		$_SESSION[self::SessionArrayVar]['SessionUserId'] = isset($sessionUser) ? $sessionUser->GetUserId() : 0;
		self::$sessionUser = $sessionUser;

		if (isset($sessionUser))
		{
			$sessionUser->UpdateLastLoginTime();
			UnbLog::Add(UnbLogType::Information, UnbLogEvent::UserLogin);
		}
	}

	// Converts a plain password for storing it in the database.
	//
	// password = (string) Plain password as entered by the user
	//
	// Returns (string) password hash
	//
	public static function ConvertPassword($password)
	{
		// Add random salt.
		// Adding 4 bytes makes the hash 36 bytes long which nicely fits into base64 encoding
		// without padding, generating 48 bytes.
		$salt = '';
		for ($i = 0; $i < 4; $i++)
			$salt .= chr(mt_rand(0, 255));
		return base64_encode($salt . hash('sha256', $salt . $password, true));
	}

	// Validates a plain password against a stored password hash.
	//
	// password = (string) Plain password as entered by the user
	// hash = (string) Password hash as stored in the database
	//
	// Returns (bool) {{true}}, if the password matches the hash, or {{false}} otherwise.
	//
	private static function ValidatePassword($password, $hash)
	{
		$hash = base64_decode($hash);
		$salt = substr($hash, 0, 4);
		return $hash === $salt . hash('sha256', $salt . $password, true);
	}

	// Validates the current session user's password.
	//
	// password = (string) Plain password as entered by the user
	//
	// Returns (bool) {{true}}, if the password is correct, or {{false}} otherwise.
	//
	public static function ValidateSessionUserPassword($password)
	{
		if (isset($password))
		{
			if (!is_string($password))
				throw new InvalidArgumentException('Invalid argument type: $password');
			$password = trim($password);
			if (!strlen($password))
				$password = null;
		}

		if (!self::IsLoggedIn())
			throw new BadMethodCallException('Not logged in');

		$sessionUserPassword = self::GetSessionUser()->GetLoginPassword();
		if (!isset($sessionUserPassword) && !isset($password))
			return true;
		return self::ValidatePassword($password, $sessionUserPassword);
	}

	// Generates a new random password.
	//
	// This password may use any printable ASCII characters.
	//
	// length = (int) Length of the generated password in characters
	//
	// Returns (string) The generated password
	//
	public static function GeneratePassword($length = 12)
	{
		$p = '';
		while ($length--)
			$p .= chr(mt_rand(0x21, 0x7E));
		return $p;
	}

	// Encrypts data with the system encrpytion key.
	//
	// This function uses Twofish encryption in ECB mode which is best suitable for short input
	// data like keys and passwords. It is not recommended for file contents. See the PHP manual
	// about the mcrypt library for details.
	//
	// data = (string) Data to encrypt
	//
	// Returns (string) Encrypted data in base64 encoding
	//
	public static function Encrypt($data)
	{
		if (!isset($data)) return null;

		$td = mcrypt_module_open(MCRYPT_TWOFISH, '', MCRYPT_MODE_ECB, '');
		$iv = mcrypt_create_iv(mcrypt_enc_get_iv_size($td), MCRYPT_RAND);
		$key = UnbConfiguration::Get('encryption.key');
		if (!isset($key) || !strlen($key))
			$key = self::DefaultEncryptionKey;
		$key = substr($key, 0, mcrypt_enc_get_key_size($td));
		mcrypt_generic_init($td, $key, $iv);
		$encrypted_data = mcrypt_generic($td, $data);
		mcrypt_generic_deinit($td);
		mcrypt_module_close($td);
		return base64_encode($encrypted_data);
	}

	// Decrypts data with the system encrpytion key.
	//
	// data = (string) Data to decrypt, in base64 encoding
	//
	// Returns (string) Decrypted data
	//
	public static function Decrypt($data)
	{
		if (!isset($data)) return null;

		$td = mcrypt_module_open(MCRYPT_TWOFISH, '', MCRYPT_MODE_ECB, '');
		$iv = mcrypt_create_iv(mcrypt_enc_get_iv_size($td), MCRYPT_RAND);
		$key = UnbConfiguration::Get('encryption.key');
		if (!isset($key) || !strlen($key))
			$key = self::DefaultEncryptionKey;
		$key = substr($key, 0, mcrypt_enc_get_key_size($td));
		mcrypt_generic_init($td, $key, $iv);
		$decrypted_data = mdecrypt_generic($td, base64_decode($data));
		mcrypt_generic_deinit($td);
		mcrypt_module_close($td);
		return rtrim($decrypted_data, "\0");   // Remove padding, see http://php.net/mdecrypt_generic
	}

	// Gets a URL key that identifies a request as valid for the current session.
	//
	// Returns (string) URL key
	//
	public static function GetUrlKey()
	{
		if (!session_id())
			throw new BadMethodCallException('Session not enabled');

		return sprintf('%08X', crc32(session_id()));
	}

	// Checks a given URL key.
	//
	// key = (string) URL key value as returned from GetUrlKey()
	//
	// Returns (bool) {{true}} if the key is accepted, {{false}} otherwise
	//
	public static function CheckUrlKey($key)
	{
		if (!session_id())
			throw new BadMethodCallException('Session not enabled');

		return $key === sprintf('%08X', crc32(session_id()));
	}

} // class UnbSession

?>
