<?php
// UNB Components
// Copyright since 2007 by Yves Goergen
// Website: http://unclassified.de/go/unb2
// Licence: GNU General Public License (GPL) Version 3
//
// This file contains the UnbUser class.

if (!class_exists('UnbDatabase', false)) require_once dirname(__FILE__) . '/database.class.php';
if (!class_exists('UnbSession', false)) require_once dirname(__FILE__) . '/session.class.php';
if (!class_exists('UnbSecurityException', false)) require_once dirname(__FILE__) . '/unb.class.php';

// Enumeration of vcard entry visibility values.
//
class UnbUserVcardVisibility
{
	// Visible for all users
	const Public_ = 0;
	// Visible only to registered and logged in users
	const Internal = 1;
	// Visible only to trusted users
	const Confidential = 2;
	// Visible only to the owner user
	const Private_ = 3;
} // class UnbUserVcardVisibility

// Enumeration of standard keys.
//
class UnbUserKey
{
	// The system key for the system itself
	const System = 1;
	// The system key for administrative users
	const Administrator = 2;
	// The system key for moderator users
	const Moderator = 3;
	// The system key for trusted users
	const Trusted = 4;
	// The system key for a virtual default user, used for default user configuration settings.
	// This key must not be assigned to any user as additional key.
	const DefaultUser = 5;
} // class UnbUserKey

// The user class represents a single user in the UNB system and provides methods to find and
// manipulate user instances.
//
// It depends on the UnbDatabase class for a database connection.
//
class UnbUser extends UnbEntity
{
	// Length of new user ID numbers (decimal digits). This must be in the range of 2...9.
	// The key length is used to generate new user IDs. Any length of user IDs is compatible,
	// so the key length can be changed at any time. When you're running out of keys, increasing
	// the key length is recommended. Longer keys have no worse performance than shorter keys,
	// they just... "look longer".
	const KeyLength = 5;

	// Determines whether a secure password is required for users.
	const RequireSecurePassword = false;

	// User ID, also called "key"
	private $userId = 0;
	// Time when this user was created
	private $createdTime = '';
	// Time when this user has last logged in
	private $lastLoginTime = '';
	// Display name, used to identify user keys
	private $displayName = '';
	// Login name
	private $loginName = '';
	// Login password
	private $loginPassword = '';

	// TODO: Distinguish between "personal" and "virtual" keys:
	// Personal keys must not be part of a user's additional keys list.
	// Virtual keys must not have an additional keys list on their own. (And they must not have logon information.)

	// Cached additional keys list
	private $additionalKeys = null;

	// Cached UnbDatabase instance
	private $db;

	private function __construct()
	{
		$this->db = UnbDatabase::GetInstance();
	}

	// Gets an array of keys that must not be added to a key list.
	//
	// This function is used from the key list management functions in UnbDatabase.
	//
	// Returns (array(int)) List of keys
	//
	public static function GetNonListKeys()
	{
		return array(UnbUserKey::System, UnbUserKey::DefaultUser);
	}

	// Creates a new user instance.
	//
	// The newly created user instance is also stored in the database.
	//
	// If a user ID is automatically generated, a random number of KeyLength decimal digits is
	// generated. If the new ID is not unique, it will be retried 50 times before the process
	// fails.
	//
	// displayName = (string) New user's display name or {{null}} (max. 255 characters)
	// loginName = (string) New user's login name or {{null}} (max. 255 characters)
	// loginPassword = (string) New user's login password or {{null}} (used with loginName)
	// newId = (int) Use this ID for the new user instead of automatically generating one.
	//               This is meant to be used for setup and maintenance only!
	// copyDefaultSettings = (bool) Copy the default user settings to the newly created user
	//
	// Throws InvalidArgumentException
	// Throws LengthException
	// Throws UnbSecurityException if the display or login name are already assigned.
	// Throws UnbDatabaseException
	//
	// Uses UnbUser::KeyLength
	//
	// Returns the newly created UnbUser object
	//
	public static function Create($displayName = null, $loginName = null, $loginPassword = null, $newId = null, $copyDefaultSettings = false)
	{
		// Validate data and configuration
		if (isset($newId))
		{
			if (!is_int($newId))
				throw new InvalidArgumentException('Invalid argument type: $newId');
			if ($newId <= 0)
				throw new InvalidArgumentException('Argument out of range: $newId');
		}
		else
		{
			if (self::KeyLength < 2 || self::KeyLength > 9)
				throw new InvalidArgumentException('Argument out of range: KeyLength');
		}
		if (isset($displayName))
		{
			if (!is_string($displayName))
				throw new InvalidArgumentException('Invalid argument type: $displayName');
			$displayName = trim($displayName);
			if (mb_strlen($displayName) > 255)
				throw new LengthException('[UNB#16] Display name is too long');
			if (!strlen($displayName))
				$displayName = null;
		}
		if (isset($loginName))
		{
			if (!is_string($loginName))
				throw new InvalidArgumentException('Invalid argument type: $loginName');
			$loginName = mb_strtolower(trim($loginName));
			if (mb_strlen($loginName) > 255)
				throw new LengthException('[UNB#17] Login name is too long');
			if (!strlen($loginName))
				$loginName = null;
		}
		if (isset($loginPassword))
		{
			if (!is_string($loginPassword))
				throw new InvalidArgumentException('Invalid argument type: $loginPassword');
			$loginPassword = trim($loginPassword);
			if (!strlen($loginPassword))
				$loginPassword = null;
		}
		if (isset($loginName) ^ isset($loginPassword))
			throw new InvalidArgumentException('[UNB#18] Either both login name and password or none at all must be set.');
		if (self::RequireSecurePassword && isset($loginPassword) && !self::IsSecurePassword($loginPassword, $loginName, $err))
			throw new UnbSecurityException('[UNB#19] Login password is not considered secure. ' . $err);

		$displayNamePrep = null;
		if (isset($displayName))
			$displayNamePrep = self::PrepDisplayName($displayName);

		$now = time();

		$db = UnbDatabase::GetInstance();
		$db->StartTransaction();
		try
		{
			// Loop to allow to repeat the ID selection process if it should fail in the very last
			// moment, ie. when someone else had the same idea and created the selected ID a
			// moment before. Normally, this loop is executed just once then left. To avoid an
			// endless loop in case of a permanent problem, the retry counter is set to a low
			// value.
			$selectionRetries = 3;
			while (true)
			{
				if (isset($newId))
				{
					$tmpNewId = $newId;
				}
				else
				{
					// Generate a new unique user ID
					$ps = $db->Prepare('SELECT COUNT(*) FROM "#user" WHERE "UserId" = ?');
					$retries = 20;
					$minId = pow(10, self::KeyLength - 1);
					$maxId = pow(10, self::KeyLength) - 1;
					do
					{
						// Find a new random number to start with
						$tmpNewId = mt_rand($minId, $maxId);
						if ($retries == 1)
						{
							// At the very last try, scan the entire defined number range
							// (In a test, this has first occured at 99% ID space usage.)
							$tmpNewId = $minId;
						}

						$count = $db->QueryValue($ps, $tmpNewId);

						if ($count > 0 && $retries <= 3)
						{
							// Near the end, try scanning for a free ID from the random number upwards.
							// In tests, this has proven to be an efficient method to find free IDs
							// when only few free IDs are available. This query does not take
							// considerably longer than a single ID try.
							// (In a test, this has first occured at ~70% ID space usage.)
							$tmpNewId = $db->QueryValue(
								'SELECT MIN("UserId" + 1) AS "FreeId"
								FROM "#user" u1
								WHERE NOT EXISTS (SELECT * FROM "#user" u2 WHERE u2."UserId" = u1."UserId" + 1)
									AND "UserId" >= ?',
								$tmpNewId);
							if ($tmpNewId <= $maxId)
							{
								// The new ID is within the defined range, accept it
								$count = 0;
							}
						}

						$retries--;
					}
					while ($count > 0 && $retries > 0);

					if ($count > 0)
					{
						throw new Exception('[UNB#22] No free user ID available. Consider raising the key length.');
					}
				}

				$fields = '"UserId", "CreatedTime"';
				$values = '?, ?';
				$parameters = array(
					$tmpNewId,
					UnbDatabase::ConvertDateTime($now));
				if (isset($displayName))
				{
					$fields .= ', "DisplayName", "DisplayNamePrep"';
					$values .= ', ?, ?';
					$parameters[] = $displayName;
					$parameters[] = $displayNamePrep;
				}
				if (isset($loginName))
				{
					$fields .= ', "LoginName"';
					$values .= ', ?';
					$parameters[] = $loginName;
				}
				if (isset($loginPassword))
				{
					$loginPassword = UnbSession::ConvertPassword($loginPassword);
					$fields .= ', "LoginPassword"';
					$values .= ', ?';
					$parameters[] = $loginPassword;
				}
				$sql = 'INSERT INTO "#user" (' . $fields . ') VALUES (' . $values . ')';

				try
				{
					$db->Execute($sql, $parameters);
				}
				catch (UnbDatabaseException $ex)
				{
					if ($db->IsConstraintError())
					{
						if (strpos($db->GetErrorMessage(), 'PRIMARY') !== false && $selectionRetries-- > 0)
							// Repeat the ID selection, see above.
							continue;
						if (strpos($db->GetErrorMessage(), 'user_DisplayNamePrep_unq') !== false)
							throw new UnbSecurityException('[UNB#20] Display name is already assigned.');
						if (strpos($db->GetErrorMessage(), 'user_LoginName_unq') !== false)
							throw new UnbSecurityException('[UNB#21] Login name is already assigned.');
					}
					throw $ex;
				}

				// Leave the loop, the selected ID was still free.
				break;
			}

			$user = new self();
			$user->userId = $tmpNewId;
			$user->createdTime = $now;
			$user->displayName = $displayName;
			$user->loginName = $loginName;
			$user->loginPassword = $loginPassword;

			if ($copyDefaultSettings)
			{
				$defaultUser = self::FindByUserId(UnbUserKey::DefaultUser);
				$defaultSettings = $defaultUser->GetSettings();
				$user->SetSetting($defaultSettings);
			}

			$db->Commit();
			return $user;
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}
	}

	// Creates a UnbUser object from a database record array.
	//
	// rec = (array) Database record
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	//
	// Returns (UnbUser) object constructed from the database record
	//
	protected static function FromRecord($rec)
	{
		if (!is_array($rec))
			throw new InvalidArgumentException('Invalid argument type: $rec');
		if (!count($rec))
			throw new InvalidArgumentException('Record array is empty');

		$user = new self();
		$user->userId = intval($rec['UserId']);
		$user->displayName = $rec['DisplayName'];
		$user->createdTime = UnbDatabase::ConvertDateTime($rec['CreatedTime']);
		$user->lastLoginTime = UnbDatabase::ConvertDateTime($rec['LastLoginTime']);
		$user->loginName = $rec['LoginName'];
		$user->loginPassword = $rec['LoginPassword'];
		return $user;
	}

	// Finds a user instance by its user ID.
	//
	// userId = (int) User ID
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbDatabaseException
	//
	// Returns a UnbUser object, if successful, {{null}} otherwise.
	//
	public static function FindByUserId($userId)
	{
		if (!is_int($userId))
			throw new InvalidArgumentException('Invalid argument type: $userId');
		if ($userId <= 0)
			throw new InvalidArgumentException('Argument out of range: $userId');

		$db = UnbDatabase::GetInstance();
		$rec = $db->QueryRecord('SELECT * FROM "#user" WHERE "UserId" = ?', $userId);
		if (!$rec) return null;
		return self::FromRecord($rec);
	}

	// Finds a user instance by its display name.
	//
	// displayName = (string) Display name, searched case-insensitive (max. 255 characters)
	//
	// Throws InvalidArgumentException
	// Throws LengthException
	// Throws UnbDatabaseException
	//
	// Returns a UnbUser object, if successful, {{null}} otherwise.
	//
	public static function FindByDisplayName($displayName)
	{
		if (!is_string($displayName))
			throw new InvalidArgumentException('Invalid argument type: $displayName');
		$displayName = trim($displayName);
		if (!strlen($displayName))
			throw new InvalidArgumentException('Display name is empty');
		if (mb_strlen($displayName) > 255)
			throw new LengthException('[UNB#16] Display name is too long');

		$displayNamePrep = self::PrepDisplayName($displayName);

		$db = UnbDatabase::GetInstance();
		$rec = $db->QueryRecord(
			'SELECT * FROM "#user" WHERE "DisplayNamePrep" = ?',
			$displayNamePrep);
		if (!$rec) return null;
		return self::FromRecord($rec);
	}

	// Finds a user instance by its login name.
	//
	// loginName = (string) Login name, searched case-insensitive (max. 255 characters)
	//
	// Throws InvalidArgumentException
	// Throws LengthException
	// Throws UnbDatabaseException
	//
	// Returns a UnbUser object, if successful, {{null}} otherwise.
	//
	public static function FindByLoginName($loginName)
	{
		if (!is_string($loginName))
			throw new InvalidArgumentException('Invalid argument type: $loginName');
		$loginName = mb_strtolower(trim($loginName));
		if (!strlen($loginName))
			throw new InvalidArgumentException('Login name is empty');
		if (mb_strlen($loginName) > 255)
			throw new LengthException('[UNB#17] Login name is too long');

		$db = UnbDatabase::GetInstance();
		$rec = $db->QueryRecord(
			'SELECT * FROM "#user" WHERE LOWER("LoginName") = LOWER(?)',
			$loginName);
		if (!$rec) return null;
		return self::FromRecord($rec);
	}

	// Finds all non-system user accounts.
	//
	// Throws UnbDatabaseException
	//
	// Returns UnbUser objects for the found users.
	//
	// TODO: Add arguments to narrow down the search.
	//
	public static function Find()
	{
		$db = UnbDatabase::GetInstance();
		$recs = $db->Query('SELECT * FROM "#user" WHERE "UserId" > 9 ORDER BY "DisplayNamePrep"');
		if (!isset($recs)) return null;
		$users = array();
		foreach ($recs as $rec)
		{
			$users[] = self::FromRecord($rec);
		}
		return $users;
	}

	// Validates a password against the secure password policy.
	//
	// password = (string) The password to be checked
	// data = (string) Additional data, usually the login name
	// out err = (string) Error message in case of failure
	//
	// Returns (bool) {{true}} if the password is secure, {{false}} otherwise
	//
	public static function IsSecurePassword($password, $data, &$err)
	{
		$err = '';

		if (mb_strlen($password) < 6)
		{
			$err = 'Require at least 6 characters';
			#$err = 'Mindestens 6 Zeichen erforderlich';
			return false;
		}
		if (mb_strtolower($password) == mb_strtolower($data))
		{
			$err = 'Reject password caseless equal to username';
			#$err = 'Kennwort ist gleich dem Benutzernamen (caseless)';
			return false;
		}
		if ($data != '' && mb_strpos(mb_strtolower($password), mb_strtolower($data)) !== false)
		{
			$err = 'Reject password caseless containing username';
			#$err = 'Kennwort enthlt den Benutzernamen (caseless)';
			return false;
		}
		if (preg_match('/[^!-~]/', $password))
		{
			$err = 'Reject non-ASCII characters';
			#$err = 'Nicht-ASCII-Zeichen sind unzulssig';
			return false;
		}
		if (!(preg_match('/[a-z]/', $password) && preg_match('/[0-9]/', $password) ||
			  preg_match('/[a-z]/', $password) && preg_match('/[A-Z]/', $password) ||
			  preg_match('/[A-Z]/', $password) && preg_match('/[0-9]/', $password) ||
			  preg_match('/[0-9]/', $password) && preg_match('/[!#$%&()*+,\-.\/:;<=>?@\[\\\\\]^_|~]/', $password) ||
			  preg_match('/[a-z]/', $password) && preg_match('/[!#$%&()*+,\-.\/:;<=>?@\[\\\\\]^_|~]/', $password) ||
			  preg_match('/[A-Z]/', $password) && preg_match('/[!#$%&()*+,\-.\/:;<=>?@\[\\\\\]^_|~]/', $password)))
		{
			$err = 'Require at least 2 character classes out of numbers, lowercase, uppercase characters and other symbols';
			#$err = 'Mindestens 2 Zeichenklassen aus Zahlen, Kleinbuchstaben, Grobuchstaben und anderen Zeichen erforderlich';
			return false;
		}
		if (preg_match('/(.)\1\1/', $password))
		{
			$err = 'Reject the same character 3 times or more';
			#$err = 'Kein Zeichen darf 3 mal oder fter wiederholt werden';
			return false;
		}
		for ($n = 0; $n <= 7; $n++)
			if (preg_match('/' . $n . ($n + 1) . ($n + 2) . '/', $password))
			{
				$err = 'Reject simple sequences: 012, 123, ...';
				#$err = 'Einfache Folgen wie 012, 123, ... sind unzulssig';
				return false;
			}
		for ($n = 0x41; $n <= 0x58; $n++)
			if (preg_match('/' . chr($n) . chr($n + 1) . chr($n + 2) . '/', $password))
			{
				$err = 'Reject simple sequences: ABC, BCD, ...';
				#$err = 'Einfache Folgen wie ABC, BCD, ... sind unzulssig';
				return false;
			}
		for ($n = 0x61; $n <= 0x78; $n++)
			if (preg_match('/' . chr($n) . chr($n + 1) . chr($n + 2) . '/', $password))
			{
				$err = 'Reject simple sequences: abc, bcd, ...';
				#$err = 'Einfache Folgen wie abc, bcd, ... sind unzulssig';
				return false;
			}
		// TODO: count_chars() is not exactly UTF-8 safe
		if (strlen(count_chars($password, 3)) < 4)
		{
			$err = 'Require at least 4 different characters to be used';
			#$err = 'Mindestens 4 unterschiedliche Zeichen erforderlich';
			return false;
		}

		return true;
	}

	// Gets the user ID.
	//
	// Returns (int) user ID
	//
	public function GetUserId()
	{
		return $this->userId;
	}

	// Gets the time when the user was created.
	//
	// Returns (int) UNIX timestamp
	//
	public function GetCreatedTime()
	{
		return $this->createdTime;
	}

	// Gets the time when the user has last logged in.
	//
	// Returns (int) UNIX timestamp
	//
	public function GetLastLoginTime()
	{
		return $this->lastLoginTime;
	}

	// Sets the last login time to now.
	//
	public function UpdateLastLoginTime()
	{
		if (UnbSession::GetSessionUserId() != $this->GetUserId())
			throw new UnbSecurityException('You must be logged in as this user to set the last login time.');

		$now = time();

		$this->db->StartTransaction();
		try
		{
			$this->db->Execute(
				'UPDATE "#user" SET "LastLoginTime" = ? WHERE "UserId" = ?',
				array(UnbDatabase::ConvertDateTime($now), $this->userId));
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
		$this->lastLoginTime = $now;
	}

	// Gets the display name.
	//
	// Returns (string) display name
	//
	public function GetDisplayName()
	{
		return $this->displayName;
	}

	// Gets the display name, if set, or the user ID.
	//
	// Returns (string) display name or user ID
	//
	public function GetDisplayNameOrId()
	{
		return isset($this->displayName) && strlen($this->displayName) ? $this->displayName : $this->userId;
	}

	// Sets the display name and stores it in the database.
	//
	// displayName = (string) New display name or {{null}} (max. 255 characters)
	//
	// Throws InvalidArgumentException
	// Throws LengthException
	// Throws UnbSecurityException if not logged in or not the same user, or if the display name is already assigned.
	// Throws UnbDatabaseException
	//
	// Privileges: Must be logged in as the same user
	//
	public function SetDisplayName($displayName)
	{
		if (!is_string($displayName))
			throw new InvalidArgumentException('Invalid argument type: $displayName');
		$displayName = trim($displayName);
		if (!strlen($displayName))
			throw new InvalidArgumentException('Display name is empty');
		if (mb_strlen($displayName) > 255)
			throw new LengthException('[UNB#16] Display name is too long');

		if (UnbSession::GetSessionUserId() != $this->userId)
			throw new UnbSecurityException('You must be logged in as this user to set the display name.');
		// TODO: Why can't administrators do that, too?

		$displayNamePrep = self::PrepDisplayName($displayName);

		$this->db->StartTransaction();
		try
		{
			$this->db->Execute(
				'UPDATE "#user" SET "DisplayName" = ?, "DisplayNamePrep" = ? WHERE "UserId" = ?',
				array($displayName, $displayNamePrep, $this->userId));
			$this->db->Commit();
		}
		catch (UnbDatabaseException $ex)
		{
			$this->db->RollBack();
			if ($db->IsConstraintError())
			{
				if (strpos($db->GetErrorMessage(), 'user_DisplayNamePrep_unq') !== false)
					throw new UnbSecurityException('[UNB#20] Display name is already assigned.');
			}
			throw $ex;
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
		$this->displayName = $displayName;
	}

	// Prepares the specified display name for use in a uniqueness test or for sorting.
	//
	// displayName = (string) Display name to be prepared
	//
	// Returns (string) Prepared display name
	//
	public static function PrepDisplayName($displayName)
	{
		return mb_strtolower($displayName);
	}

	// Gets the login name.
	//
	// Returns (string) login name
	//
	// Throws UnbSecurityException if not logged in or not the same user/no administrator.
	//
	// Privileges: Must be logged in as the same user or with administrator privilege
	//
	public function GetLoginName()
	{
		if (!UnbSession::IsAdministrator() && UnbSession::GetSessionUserId() != $this->userId)
			throw new UnbSecurityException('You must be logged in as this user or need the administrator privilege to get a login name.');

		return $this->loginName;
	}

	// Sets the login name and stores it in the database.
	//
	// loginName = (string) New login name, stored in lower case, or {{null}} (max. 255 characters)
	//
	// Throws InvalidArgumentException
	// Throws LengthException
	// Throws UnbSecurityException if not logged in or not the same user/no administrator, or if the login name is already assigned.
	// Throws UnbDatabaseException
	//
	// Privileges: Must be logged in as the same user or with administrator privilege
	//
	public function SetLoginName($loginName)
	{
		if (isset($loginName))
		{
			if (!is_string($loginName))
				throw new InvalidArgumentException('Invalid argument type: $loginName');
			$loginName = mb_strtolower(trim($loginName));
			if (mb_strlen($loginName) > 255)
				throw new LengthException('[UNB#17] Login name is too long');
			if (!strlen($loginName))
				$loginName = null;
		}

		if (!UnbSession::IsAdministrator() && UnbSession::GetSessionUserId() != $this->userId)
			throw new UnbSecurityException('You must be logged in as this user or need the administrator privilege to set a login name.');

		$this->db->StartTransaction();
		try
		{
			$this->db->Execute(
				'UPDATE "#user" SET "LoginName" = ? WHERE "UserId" = ?',
				array($loginName, $this->userId));
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			if ($this->db->IsConstraintError())
				throw new UnbSecurityException('[UNB#21] Login name is already assigned.');
			throw $ex;
		}
		$this->loginName = $loginName;
	}

	// Gets the login password hash.
	//
	// Returns (string) login password hash
	//
	public function GetLoginPassword()
	{
		return $this->loginPassword;
	}

	// Sets the login password and stores it in the database.
	//
	// loginPassword = (string) New plain password, as entered by the user
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbSecurityException if not logged in or not the same user/no administrator.
	// Throws UnbDatabaseException
	//
	// Privileges: Must be logged in as the same user or with administrator privilege
	//
	public function SetLoginPassword($loginPassword)
	{
		if (isset($loginPassword))
		{
			if (!is_string($loginPassword))
				throw new InvalidArgumentException('Invalid argument type: $loginPassword');
			$loginPassword = trim($loginPassword);
			if (!strlen($loginPassword))
				$loginPassword = null;
		}
		if (isset($loginPassword))
		{
			if (self::RequireSecurePassword && !self::IsSecurePassword($loginPassword, $this->GetLoginName(), $err))
				throw new UnbSecurityException('[UNB#19] Login password is not considered secure. ' . $err);
			// Convert the plain password into a hash to be stored in the database.
			$loginPassword = UnbSession::ConvertPassword($loginPassword);
		}

		if (!UnbSession::IsAdministrator() && UnbSession::GetSessionUserId() != $this->userId)
			throw new UnbSecurityException('You must be logged in as this user or need the administrator privilege to set a login password.');

		$this->db->StartTransaction();
		try
		{
			$this->db->Execute(
				'UPDATE "#user" SET "LoginPassword" = ? WHERE "UserId" = ?',
				array($loginPassword, $this->userId));
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
		$this->loginPassword = $loginPassword;
	}

	// Gets a value indicating whether this user is an administrator.
	//
	// Throws UnbDatabaseException
	//
	// Returns (bool)
	//
	public function IsAdministrator()
	{
		if ($this->userId == UnbUserKey::Administrator) return true;
		$keys = $this->additionalKeys;
		if (!isset($keys))
			$keys = $this->GetAdditionalKeys();
		return in_array(UnbUserKey::Administrator, $keys);
	}

	// Gets a value indicating whether this user is a moderator.
	//
	// Throws UnbDatabaseException
	//
	// Returns (bool)
	//
	public function IsModerator()
	{
		if ($this->userId == UnbUserKey::Moderator) return true;
		$keys = $this->additionalKeys;
		if (!isset($keys))
			$keys = $this->GetAdditionalKeys();
		return in_array(UnbUserKey::Moderator, $keys);
	}

	// Gets a value indicating whether this user is called "trusted".
	//
	// Throws UnbDatabaseException
	//
	// Returns (bool)
	//
	public function IsTrusted()
	{
		if ($this->userId == UnbUserKey::Trusted) return true;
		$keys = $this->additionalKeys;
		if (!isset($keys))
			$keys = $this->GetAdditionalKeys();
		return in_array(UnbUserKey::Trusted, $keys);
	}

	// Gets all additional keys assigned to this user.
	//
	// Throws UnbDatabaseException
	//
	// Returns (array(int)) keys array
	//
	public function GetAdditionalKeys()
	{
		$this->additionalKeys = $this->db->GetKeylistKeys('user', 'AdditionalKeylistId', 'UserId', $this->userId);
		return $this->additionalKeys;
	}

	// Assigns an additional key to this user.
	//
	// key = (int/UnbUser) Key to add
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbSecurityException if not logged in or no administrator.
	// Throws UnbDatabaseException
	//
	// See UnbDatabase::AddKeylistKey()
	//
	// Privileges: Must be logged in with administrator privilege
	//
	public function AddAdditionalKey($key)
	{
		if (!UnbSession::IsAdministrator())
			throw new UnbSecurityException('You need the administrator privilege to add additional user keys.');
		// TODO: Maybe moderators may add the UnbUserKey::Trusted to a user

		$this->additionalKeys = null;   // Clear cache
		$this->db->AddKeylistKey('#user', 'AdditionalKeylistId', 'UserId', $this->userId, $key);
	}

	// Divests an additional key from this user.
	//
	// key = (int/UnbUser) Key to remove
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbSecurityException if not logged in or no administrator.
	// Throws UnbDatabaseException
	//
	// See UnbDatabase::RemoveKeylistKey()
	//
	// Privileges: Must be logged in with administrator privilege
	//
	public function RemoveAdditionalKey($key)
	{
		if (!UnbSession::IsAdministrator())
			throw new UnbSecurityException('You need the administrator privilege to remove additional user keys.');
		// TODO: Maybe moderators may remove the UnbUserKey::Trusted from a user

		$this->additionalKeys = null;   // Clear cache
		$this->db->RemoveKeylistKey('#user', 'AdditionalKeylistId', 'UserId', $this->userId, $key);
	}

	// Sets all additional keys for this user. This removes any keys set before.
	//
	// keys = (array(int/UnbUser)) Keys to set
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbSecurityException if not logged in or no administrator.
	// Throws UnbDatabaseException
	//
	// See UnbDatabase::ResetKeylistKeys()
	//
	// Privileges: Must be logged in with administrator privilege
	//
	public function SetAdditionalKeys($keys)
	{
		if (!UnbSession::IsAdministrator())
			throw new UnbSecurityException('You need the administrator privilege to set additional keys.');

		$this->additionalKeys = null;   // Clear cache
		$this->db->ResetKeylistKeys('#user', 'AdditionalKeylistId', 'UserId', $this->userId, $keys);
	}

	// Deletes this instance of the user with all its references.
	//
	// Note that deleting a user which is the author of messages or message revisions is not possible.
	//
	// Throws UnbSecurityException if not logged in or no administrator.
	// Throws UnbDatabaseException
	//
	// Privileges: Must be logged in with administrator privilege
	//
	public function Delete()
	{
		if (!UnbSession::IsAdministrator())
			throw new UnbSecurityException('You need the administrator privilege to delete a user.');

		$this->additionalKeys = null;   // Clear cache
		$this->db->StartTransaction(true);
		try
		{
			// Deleting users may leave keylists unreferenced. Delete them now.
			$this->db->Execute(
				'DELETE FROM "#keylist"
				WHERE "KeylistId" =
					(SELECT "AdditionalKeylistId"
					FROM "#user"
					WHERE "UserId" = ?)',
				$this->userId);

			// Keys lists that only contain the user to be deleted would be empty after deleting
			// the user. In cases where the key list is restricting access to something, the old
			// user ID must be replaced by the system key to prevent dropping the access
			// restriction. For other cases the key list may be dropped altogether.

			// First, find all keylists that are referenced as an access-restricting key list and
			// that only contain the user to be deleted. Also, lock the read key list IDs while
			// we're working with them.
			$keylistIds = $this->db->QueryArray(
				'SELECT "KeylistId"
				FROM "#keylist" k
				WHERE
					-- KeylistId is referenced as an access-restricting key list
					("KeylistId" IN (SELECT "ReadAccessKeylistId" FROM "#message") OR
					"KeylistId" IN (SELECT "AlterAccessKeylistId" FROM "#message") OR
					"KeylistId" IN (SELECT "ReplyAccessKeylistId" FROM "#message") OR
					"KeylistId" IN (SELECT "UseAccessKeylistId" FROM "#tag") OR
					"KeylistId" IN (SELECT "ReadAccessKeylistId" FROM "#tag")) AND
					-- The key list does not contain any other UserId than that of the user to be deleted
					NOT EXISTS (
						SELECT *
						FROM "#keylist_keys" kk
						WHERE kk."KeylistId" = k."KeylistId" AND
							kk."UserId" <> ?)
				FOR UPDATE',
				$this->userId);
			$keylistIds = array_map('intval', $keylistIds);

			// Now replace the old user ID with the system key in all affected key lists
			if (count($keylistIds))
			{
				$this->db->Execute(
					'UPDATE "#keylist_keys"
					SET "UserId" = ?
					WHERE "UserId" = ? AND
						"KeylistId" IN (' . join(', ', $keylistIds) . ')',
					array(UnbUserKey::System, $this->userId));

				// NOTE:
				// MySQL doesn't currently support SELECTing from a table being UPDATEd. Sad but true.
				// This is why the list of KeylistIds is first retrieved and then used in the UPDATE
				// statement as literal.
				// See http://dev.mysql.com/doc/refman/5.1/en/update.html (at the very bottom)
			}

			// When the user is deleted, all its remaining occurences in key lists are removed
			// (cascading foreign key reference)
			$this->db->Execute('DELETE FROM "#user" WHERE "UserId" = ?', $this->userId);

			// All remaining references to now empty key lists can now be NULLed. A database
			// trigger already did that for us. :)

			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
	}

	// Gets all or the selected vcard entries for this user.
	//
	// types = (array) Vcard types to retrieve from the database, {{null}} to retrieve all types.
	//         (string) A single type name can be passed as string (max. 255 characters)
	//
	// Throws InvalidArgumentException
	// Throws LengthException
	// Throws UnbDatabaseException
	//
	// Returns an array(Type => array('Data' => ..., 'Visibility' => ...) of the requested entries.
	// Example: To get the e-mail value, use $entries['EMail']['Data'].
	// For the meaning of the 'Visibility' item, see the UnbUserVcardVisibility constants.
	//
	public function GetVcardEntries($types = null)
	{
		if (isset($types))
		{
			if (is_string($types))
				$types = array($types);
			if (!is_array($types))
				throw new InvalidArgumentException('Invalid argument type: $types');
			if (!count($types))
				throw new InvalidArgumentException('Types array is empty');
		}
		if (isset($visibility))
		{
			if (!is_int($visibility))
				throw new InvalidArgumentException('Invalid argument type: $visibility');
			if ($visibility < 0)
				throw new InvalidArgumentException('Argument out of range: $visibility');
		}

		if (UnbSession::GetSessionUserId() == $this->userId)
			$visibility = UnbUserVcardVisibility::Private_;   // We're the owner so we can see everything
		else if (UnbSession::IsTrusted() || UnbSession::IsAdministrator())
			$visibility = UnbUserVcardVisibility::Confidential;   // We're a trusted user or admin so we can see up to confidential visibility
		else if (UnbSession::IsLoggedIn())
			$visibility = UnbUserVcardVisibility::Internal;   // We're a registered user so we can see up to internal visibility
		else
			$visibility = UnbUserVcardVisibility::Public_;   // We're anonymous so we can only see public entries

		$sql = 'SELECT uv."Type", uv."Data", uv."Visibility"
			FROM "#user" u
				JOIN "#user_vcard" uv USING ("UserId")
			WHERE u."UserId" = ?';
		$parameters = array($this->userId);
		if (isset($types))
		{
			$sql .= ' AND (';
			for ($i = 0; $i < count($types); $i++)
			{
				if (mb_strlen($types[$i]) > 255)
					throw new LengthException('Type is too long');
				$sql .= ($i ? ' OR' : '') . ' uv."Type" = ?';
				$parameters[] = $types[$i];
			}
			$sql .= ')';
		}
		if (isset($visibility))
		{
			$sql .= ' AND uv."Visibility" <= ?';
			$parameters[] = $visibility;
		}
		$recs = $this->db->Query($sql, $parameters);
		$entries = array();
		if (is_array($recs))
		{
			foreach ($recs as $rec)
			{
				$entries[$rec['Type']] = array('Data' => $rec['Data'], 'Visibility' => $rec['Visibility']);
			}
		}
		return $entries;
	}

	// Sets a vcard entry for this user.
	//
	// Vcard entries for this user will be updated, or inserted if they don't exist yet.
	//
	// type = (string) Entry type to set for this user (max. 255 characters)
	// data = (string) Entry data (max. 65,535 characters)
	// visibility = (int) For the meaning of the 'Visibility' item, see the UnbUserVcardVisibility constants.
	//                    Set this {{null}} to not change the setting for existing entries and
	//                    use UnbUserVcardVisibility::Public_ visibility for new entries.
	//
	// Throws InvalidArgumentException
	// Throws LengthException
	// Throws UnbSecurityException if not logged in or not the same user/no administrator.
	// Throws UnbDatabaseException
	//
	// Privileges: Must be logged in as the same user or with administrator privilege
	//
	public function SetVcardEntry($type, $data, $visibility = null)
	{
		if (!is_string($type))
			throw new InvalidArgumentException('Invalid argument type: $type');
		$type = trim($type);
		if (!strlen($type))
			throw new InvalidArgumentException('Type is empty');
		if (mb_strlen($type) > 255)
			throw new LengthException('[UNB#23] Type is too long');
		if (!is_string($data))
			throw new InvalidArgumentException('Invalid argument type: $data');
		$data = trim($data);
		if (mb_strlen($data) > 65535)
			throw new LengthException('[UNB#24] Data is too long');
		if (isset($visibility))
		{
			if (!is_int($visibility))
				throw new InvalidArgumentException('Invalid argument type: $visibility');
			if ($visibility < 0)
				throw new InvalidArgumentException('Argument out of range: $visibility');
		}

		if (!UnbSession::IsAdministrator() && UnbSession::GetSessionUserId() != $this->userId)
			throw new UnbSecurityException('You must be logged in as this user or need the administrator privilege to set a vcard entry.');

		$this->db->StartTransaction();
		try
		{
			if (isset($visibility))
			{
				$this->db->Execute(
					'UPDATE "#user_vcard" SET "Data" = ?, "Visibility" = ? WHERE "UserId" = ? AND "Type" = ?',
					array($data, $visibility, $this->userId, $type));
			}
			else
			{
				$this->db->Execute(
					'UPDATE "#user_vcard" SET "Data" = ? WHERE "UserId" = ? AND "Type" = ?',
					array($data, $this->userId, $type));
			}
			if ($this->db->GetRowsAffected() == 0)
			{
				// The entry doesn't exist yet, so insert it.
				try
				{
					if (!isset($visibility)) $visibility = UnbUserVcardVisibility::Public_;
					$this->db->Execute(
						'INSERT INTO "#user_vcard" ("UserId", "Type", "Data", "Visibility") VALUES (?, ?, ?, ?)',
						array($this->userId, $type, $data, $visibility));
				}
				catch (UnbDatabaseException $ex)
				{
					// If the entry already exists, it is unchanged.
					// GetRowsAffected() has returned 0 in this case although the record has been found.
					// But it has not been altered because it is unchanged.
					if (!$this->db->IsConstraintError()) throw $ex;
				}
			}
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
	}

	// Removes a vcard entry for this user.
	//
	// type = (string) Entry type to remove for this user (max. 255 characters)
	//
	// Throws InvalidArgumentException
	// Throws LengthException
	// Throws UnbSecurityException if not logged in or not the same user/no administrator.
	// Throws UnbDatabaseException
	//
	// Privileges: Must be logged in as the same user or with administrator privilege
	//
	public function RemoveVcardEntry($type)
	{
		if (!is_string($type))
			throw new InvalidArgumentException('Invalid argument type: $type');
		$type = trim($type);
		if (!strlen($type))
			throw new InvalidArgumentException('Type is empty');
		if (mb_strlen($type) > 255)
			throw new LengthException('[UNB#23] Type is too long');

		if (!UnbSession::IsAdministrator() && UnbSession::GetSessionUserId() != $this->userId)
			throw new UnbSecurityException('You must be logged in as this user or need the administrator privilege to remove a vcard entry.');

		$this->db->StartTransaction();
		try
		{
			$this->db->Execute(
				'DELETE FROM "#user_vcard" WHERE "UserId" = ? AND "Type" = ?',
				array($this->userId, $type));
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
	}

	// Gets all or the selected configuration entries for this user.
	//
	// names = (array) Setting names to retrieve from the database, {{null}} to retrieve all names.
	//         (string) A single setting name can be passed as string (max. 255 characters)
	//
	// You can get settings by wildcards: Put a * for multiple possible characters.
	//
	// Throws InvalidArgumentException
	// Throws LengthException
	// Throws UnbDatabaseException
	//
	// Returns an array(Name => Data) of the requested settings.
	//
	public function GetSettings($names = null)
	{
		if (isset($names))
		{
			if (is_string($names))
				$names = array($names);
			if (!is_array($names))
				throw new InvalidArgumentException('Invalid argument type: $names');
			if (!count($names))
				throw new InvalidArgumentException('Names array is empty');
		}

		// TODO: Only reveal these settings to the user itself? (Update comment 'throws' and 'privileges' section)
		//       Except UnbUserKey::System and UnbUserKey::DefaultUser settings which must always be available!
		//       Autologin Cookie Password is stored here and must be available before login!
		// if (UnbSession::GetSessionUserId() != $this->userId)
		// 	throw new UnbSecurityException('You must be logged in as this user to get a user setting.');

		$sql = 'SELECT "Name", "Data" FROM "#user_config" WHERE "UserId" = ?';
		$parameters = array($this->userId);
		if (isset($names))
		{
			$sql .= ' AND (';
			for ($i = 0; $i < count($names); $i++)
			{
				if (mb_strlen($names[$i]) > 255)
					throw new LengthException('[UNB#25] Name is too long');
				if ($i)
					$sql .= ' OR ';
				if (strpos($names[$i], '*') === false)
				{
					$sql .= 'LOWER("Name") = LOWER(?)';
					$parameters[] = $names[$i];
				}
				else
				{
					$sql .= 'LOWER("Name") LIKE LOWER(?)';
					$parameters[] = str_replace('*', '%', $names[$i]);
				}
			}
			$sql .= ')';
		}
		$recs = $this->db->Query($sql, $parameters);
		if (!isset($recs)) return null;
		$entries = array();
		foreach ($recs as $rec)
		{
			$entries[$rec['Name']] = $rec['Data'];
		}
		return $entries;
	}

	// Gets one single configuration entry for this user.
	//
	// names = (string) A single setting name (max. 255 characters)
	//
	// Throws InvalidArgumentException
	// Throws LengthException
	// Throws UnbDatabaseException
	//
	// Returns (string) The requested setting data
	//
	public function GetSingleSetting($name)
	{
		if (!is_string($name))
			throw new InvalidArgumentException('Invalid argument type: $name');
		$name = trim($name);
		if (!strlen($name))
			throw new InvalidArgumentException('Name is empty');

		$a = $this->GetSettings($name);
		return $a[$name];
	}

	// Sets one or multiple configuration entries for this user.
	//
	// Configuration settings for this user will be updated, or inserted if they don't exist yet.
	// If an array was specified to update multiple settings, existing settings are either updated
	// or not, they are not removed if they do not appear in the settings array.
	//
	// name = (string) Setting name to set for this user (max. 255 characters)
	//        (array(string => string)) Associative array of setting names and data
	// data = (string) Setting data (max. 255 characters)
	//
	// Throws InvalidArgumentException
	// Throws LengthException
	// Throws UnbDatabaseException
	//
	public function SetSetting($name, $data)
	{
		$pairs = array();

		if (is_array($name))
			$pairs = $name;
		else if (is_string($name) && (is_string($data) || !isset($data)))
			$pairs = array($name => $data);
		else
			throw new InvalidArgumentException('Invalid argument type.');

		$newPairs = array();
		foreach ($pairs as $name => $data)
		{
			if (!is_string($name) ||
				!(is_string($data) || !isset($data)))
				throw new InvalidArgumentException('Invalid argument types.');

			$name = trim($name);
			if (!strlen($name))
				throw new InvalidArgumentException('Name is empty');
			if (mb_strlen($name) > 255)
				throw new LengthException('[UNB#25] Name is too long');

			$data = trim($data);
			if (mb_strlen($data) > 255)
				throw new LengthException('[UNB#26] Data is too long');

			$newPairs[$name] = $data;
		}
		$pairs = $newPairs;

		// TODO: Only reveal these settings to the user itself? (Update comment 'throws' and 'privileges' section)
		// if (UnbSession::GetSessionUserId() != $this->userId)
		// 	throw new UnbSecurityException('You must be logged in as this user to set a user setting.');

		$this->db->StartTransaction();
		try
		{
			foreach ($pairs as $name => $data)
			{
				$this->db->Execute(
					'UPDATE "#user_config" SET "Data" = ? WHERE "UserId" = ? AND LOWER("Name") = LOWER(?)',
					array($data, $this->userId, $name));
				if ($this->db->GetRowsAffected() == 0)
				{
					// The entry doesn't exist yet, so insert it.
					$this->db->Execute(
						'INSERT INTO "#user_config" ("UserId", "Name", "Data") VALUES (?, LOWER(?), ?)',
						array($this->userId, $name, $data));
				}
			}
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
	}

	// Removes a configuration setting for this user.
	//
	// name = (string) Setting name to remove for this user (max. 255 characters)
	//
	// Throws InvalidArgumentException
	// Throws LengthException
	// Throws UnbDatabaseException
	//
	public function RemoveSetting($name)
	{
		if (!is_string($name))
			throw new InvalidArgumentException('Invalid argument type: $name');
		$name = trim($name);
		if (!strlen($name))
			throw new InvalidArgumentException('Name is empty');
		if (mb_strlen($name) > 255)
			throw new LengthException('[UNB#25] Name is too long');

		// TODO: Only reveal these settings to the user itself? (Update comment 'throws' and 'privileges' section)
		// if (UnbSession::GetSessionUserId() != $this->userId)
		// 	throw new UnbSecurityException('You must be logged in as this user to remove a user setting.');

		$this->db->StartTransaction();
		try
		{
			$this->db->Execute(
				'DELETE FROM "#user_config" WHERE "UserId" = ? AND "Name" = ?',
				array($this->userId, $name));
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
	}

	// Gets the overall rating of all messages from this user.
	//
	// Throws UnbDatabaseException
	//
	// Returns an array(Number of ratings, Rating weight, Weighted rating average), or {{null}} if failed.
	//
	public function GetMessagesRating()
	{
		// TODO: Now, the only way to get an over-average rating weight is by publishing well-rated
		//       messages. Maybe there should be another, managed way to preset a user's rating
		//       weight, like for the site owner or so.

		$rec = $this->db->QueryRecord(
			'SELECT COUNT(*) AS "Count", SUM(mr."Weight") AS "TotalWeight", SUM(mr."Rating" * mr."Weight") AS "WeightedSum"
			FROM "#message_rating" mr
				JOIN "#message" m USING ("MessageId")
			WHERE m."Owner" = ?',
			$this->userId);
		if (!$rec) return null;
		// Keep the total weight to 100 at least to weaken few ratings
		$totalWeight = max(100.0, floatval($rec['TotalWeight']));
		return array(
			intval($rec['Count']),
			intval($rec['TotalWeight']),
			floatval($rec['WeightedSum']) / $totalWeight);
	}

} // class UnbUser

?>
